class App {
  constructor() {
    this.data = {users: {}};
    this.eventListeners = {'users.patch': []};
    this.socket = io({transports: ['websocket'], upgrade: false});
    this.socket.on('users.patch', patch => this.usersPatchHandler(patch));
  }

  get users() {
    return this.data.users;
  }

  addEventListener(type, listener) {
    if (!(type in this.eventListeners)) {
      throw `Unknown event type: ${type}`;
    }
    this.eventListeners[type].push(listener);
  }

  flash(message, category) {
    let iconPrefix;
    let toast;
    let toastCloseActionElement;

    switch (category) {
      case 'corpus':
        iconPrefix = '<i class="left material-icons">book</i>';
        break;
      case 'error':
        iconPrefix = '<i class="error-color-text left material-icons">error</i>';
        break;
      case 'job':
        iconPrefix = '<i class="left nopaque-icons">J</i>';
        break;
      default:
        iconPrefix = '<i class="left material-icons">notifications</i>';
        break;
    }
    toast = M.toast(
      {
        html: `
          <span>${iconPrefix}${message}</span>
          <button class="btn-flat toast-action white-text" data-action="close">
            <i class="material-icons">close</i>
          </button>
        `.trim()
      }
    );
    toastCloseActionElement = toast.el.querySelector('.toast-action[data-action="close"]');
    toastCloseActionElement.addEventListener('click', () => {toast.dismiss();});
  }

  getUserById(userId) {
    return new Promise((resolve, reject) => {
      if (userId in this.data.users) {resolve(this.users[userId]);}
      this.socket.emit('users.user.get', userId, response => {
        if (response.code === 200) {
          this.data.users[userId] = response.payload;
          resolve(this.data.users[userId]);
        } else {
          reject(response);
        }
      });
    });
  }

  usersPatchHandler(patch) {
    let jobId;
    let listener;
    let match;
    let operation;
    let re;
    let relationship;
    let ressourceId;
    let userId;

    for (operation of patch.filter(operation => operation.op === 'add')) {
      re = new RegExp(`^/users/([A-Za-z0-9]*)/corpora/([A-Za-z0-9]*)/(files)`);
      if (re.test(operation.path)) {
        [match, userId, ressourceId, relationship] = operation.path.match(re);
        if (!(relationship in this.users[userId].corpora[ressourceId])) {
          this.users[userId].corpora[ressourceId][relationship] = {};
        }
        continue;
      }
      re = new RegExp(`^/users/([A-Za-z0-9]*)/jobs/([A-Za-z0-9]*)/(inputs|results)`);
      if (re.test(operation.path)) {
        [match, userId, ressourceId, relationship] = operation.path.match(re);
        if (!(relationship in this.users[userId].jobs[ressourceId])) {
          this.users[userId].jobs[ressourceId][relationship] = {};
        }
        continue;
      }
    }
    this.data = jsonpatch.apply_patch(this.data, patch);
    for (listener of this.eventListeners['users.patch']) {listener(patch);}
  }
}
